//-----------------------------------------------------------------------------
// rt_Text.cpp
// Rich Thomson
// February 25th, 2001
//
// Demonstrate text using CD3DFont, D3DXCreateText and ID3DXFont.  Which is
// displayed at any given moment can be interactively selected by the user.
//
// D3DXCreateText creates a mesh in world space from a supplied font and
// string.  The glyph string, font, and color can be interactively selected
// by the user.  The viewing frustum is sized to encompass all the text in
// world space.
//
// Screen-space text can be provided by CD3DFont or ID3DXFont.  CD3DFont is
// an unsupported class from the SDK framework in
// %DX8SDK%\samples\multimedia\Common\d3dfont.cpp.  ID3DXFont is a supported
// interface in the D3DX library.  The following table compares the two
// methods for screen-space text:
//
// Feature                 CD3DFont        ID3DXFont
//
// Z buffered text         NO              NO
// Proportional fonts      yes             yes
// ANSI text               yes             yes
// UNICODE text            NO              yes
// supports kerning        NO              yes
// GDI during draw         NO              yes
// source available        yes             NO
// supported by MS         NO              yes
// font/texture size       scales font     unknown,
//     conflicts           to fit texture  presumably similar
//
// CD3DFont is less general than ID3DXFont, but uses a different rendering
// strategy that may be faster for your application.  ID3DXFont has the
// capabilities of the GDI function DrawText, giving it the ability to
// properly handle kerning and wide characters.
//
// There are additional choices for text as well.  Scalable fonts can
// provide glyph outlines through the GDI function GetGlyphOutline which
// can be rendered as polylines.  (D3DXCreateText uses this technique to
// construct a profile which is extruded to create the mesh.)  A vector
// font, such as the public domain Hershey fonts, can also be used to draw
// text as a polyline.
//
// The Text3D sample in the SDK demonstrates how to use CD3DFont to create
// 3D text as a textured polygon.
//
// Copyright (c) 1997-2000 Microsoft Corporation. All rights reserved.
// Portions Copyright (C) 2001 Rich Thomson.  All rights reserved.
//-----------------------------------------------------------------------------

// C++ standard includes
#include <cmath>
#include <iomanip>                  // ::std io manipulators hex, setw, etc.
#include <sstream>                  // string streams
#include <vector>                   // dynamic array

// Win32 includes
#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <commdlg.h>
#include <tchar.h>

// ATL includes
#include <atlbase.h>

// DX8 includes
#include <d3dx8.h>

// D3DFrame includes
#include "D3DApp.h"
#include "D3DFont.h"
#include "D3DUtil.h"
#include "DXUtil.h"

// Sample includes
#include "resource.h"
#include "colorsel.h"               // rt_choose_color()

//-----------------------------------------------------------------------------
// some constants for this sample
const UINT INITIAL_POINT_SIZE = 14;
const float SQRT_2 = sqrtf(2.f);

//-----------------------------------------------------------------------------
// s_d3dx_vertex
//
// Structure of vertices returned by D3DXCreateText.
//
struct s_d3dx_vertex
{
    float m_pos[3];
    float m_normal[3];
};

//-----------------------------------------------------------------------------
// tstring
// tostringstream
// tostream
//
// C++ standard string library classes in the spirit of <tchar.h>
// An extended implementation of this idea for general Win32 C++ programming
// is in my RT utility library in the include file <rt/tstring.h>.
//
// Fortunately we can get what we need from the C++ standard library
// with a couple of judicious typedefs.
//
// See "The C++ Standard Library", Nicolai Josuttis, 1999 for an
// excellent reference to the C++ Standard Library, including strings,
// streams and all the STL containers (list, vector, map, etc.),
// iterators and algorithms.
//
typedef std::basic_string<TCHAR> tstring;
typedef std::basic_ostringstream<TCHAR> tostringstream;
typedef std::basic_ostream<TCHAR> tostream;

//-----------------------------------------------------------------------------
// NUM_OF
//
// expression evaluating to the number of elements in a fixed-size array,
// usually an array initialized with an aggregate initializer.  Can also
// be the number of characters in TCHARs, not bytes, with a TCHAR array.
//
#define NUM_OF(ary_) (sizeof(ary_)/sizeof(ary_[0]))

//-----------------------------------------------------------------------------
// check_hr
//
// sanity check on all calls specific to this sample
//
inline HRESULT
check_hr(LPCTSTR file, unsigned line, HRESULT hr)
{
    if (FAILED(hr))
    {
        using namespace std;
        tostringstream buff;
        TCHAR error_text[256] = { 0 };
        ::D3DXGetErrorString(hr, error_text, NUM_OF(error_text));

        // format as "file(line): message" into the debug window.
        // if you double-click on this line in the debug output window
        // of VC++6, it will open an edit window on the named file
        // and position the cursor at that line.
        buff << file << _T("(") << line << _T("): 0x") << hex << setw(8)
            << hr << _T(" = ") << error_text << endl;
        ::OutputDebugString(buff.str().c_str());

        // Put a breakpoint on this ATLASSERT and you'll hit your
        // breakpoint every time your program encounters an unexpected
        // failed HRESULT.
        ATLASSERT(false);
    }
    return hr;
}

//-----------------------------------------------------------------------------
// THR
//
// this is similar to the THR() in my RT utility library, only this
// flavor doesn't throw() an exception, it prints to the debug stream
// and ATLASSERT()s false.
//
#define THR(hr_) check_hr(_T(__FILE__), __LINE__, hr_)

//-----------------------------------------------------------------------------
// s_enum_value<Enum>
// s_rs
// s_tss
//
// struct for storing a render state and its value.  This lets us quickly
// write code to set a bundle of render states, or a bundle of texture stage
// states by declaring an array of structs on the stack with an aggregate
// initializer and then calling a function to walk over the array.
//
template <typename Enum>
struct s_enum_value
{
    Enum m_state;
    DWORD m_value;
};
typedef s_enum_value<D3DRENDERSTATETYPE> s_rs;

//-----------------------------------------------------------------------------
// set_states
//
// Calls SetRenderState() on an array of state/value pairs.
//
inline void
set_states(IDirect3DDevice8 *device, const s_rs *states, UINT num_states)
{
    for (UINT i = 0; i < num_states; i++)
    {
        THR(device->SetRenderState(states[i].m_state, states[i].m_value));
    }
}

//-----------------------------------------------------------------------------
// check_menu
//
// Makes a menu item's check mark reflect the value of a bool.
//
inline void
check_menu(HMENU menu, UINT item, bool value)
{
    ::CheckMenuItem(menu, item, value ? MF_CHECKED : MF_UNCHECKED);
}

//-----------------------------------------------------------------------------
// toggle_menu
//
// WM_COMMAND update UI helper: toggles boolean value corresponding to a
// checked/unchecked menu item, sets the menu to reflect the new state,
// and sets the handled flag to indicate we've processed this WM_COMMAND.
//
inline void
toggle_menu(bool &state, HMENU menu, UINT item)
{
    state = !state;
    check_menu(menu, item, state);
}

//-----------------------------------------------------------------------------
// destroy
//
// Delete a pointer and then set it to zero.
//
template <typename Pointer>
inline void
destroy(Pointer &ptr)
{
    delete ptr;
    ptr = 0;
}

//-------------------------------------------------------------------------
// text
//
// Helper for using CD3DFont::DrawText, which takes a non-const type
// for a const parameter.
//
inline void
text(CD3DFont *font, float x, float y, D3DCOLOR fg, LPCTSTR str)
{
    THR(font->DrawText(x, y, fg, const_cast<LPTSTR>(str)));
}
inline void
text(CD3DFont *font, float x, float y, D3DCOLOR fg, const tstring &str)
{
    THR(font->DrawText(x, y, fg, const_cast<LPTSTR>(str.c_str())));
}

//-----------------------------------------------------------------------------
// c_get_dc
//
// Obtain a DC via ::GetDC in the c'tor and release it via ::ReleaseDC in
// the d'tor.  If this were a WTL or MFC application, I'd use their wrappers
// instead.
//
class c_get_dc
{
public:
    c_get_dc(HWND window)
        : m_window(window), m_dc(::GetDC(m_window))
    {
    }
    ~c_get_dc()
    {
        ::ReleaseDC(m_window, m_dc);
    }
    operator HDC() const { return m_dc; }

private:
    HWND m_window;
    HDC m_dc;
};

//-----------------------------------------------------------------------------
// c_push_gdi<>
//
// Select a GDIOBJ into the device context in the c'tor and
// deselect the object out by selecting in the old object in the d'tor.
//
template <typename T>
class c_push_gdi
{
public:
    c_push_gdi(HDC dc, T object)
        : m_dc(dc),
        m_old_object(static_cast<T>(::SelectObject(m_dc, object)))
    {
    }
    ~c_push_gdi()
    {
        ::SelectObject(m_dc, m_old_object);
    }
private:
    HDC m_dc;
    T m_old_object;
};

//-----------------------------------------------------------------------------
// mesh_vb_lock<>
//
// Locks a mesh's vertex buffer in its c'tor,
// Unlocks the vertex buffer in its d'tor.
// Provides a typed pointer to the vertex data.
//
template <typename Vertex>
class mesh_vb_lock
{
public:
    mesh_vb_lock(ID3DXMesh *mesh, DWORD flags = 0)
            : m_mesh(mesh),
            m_data(NULL)
    {
        THR(mesh->LockVertexBuffer(flags, reinterpret_cast<BYTE **>(&m_data)));
    }
    ~mesh_vb_lock()
    {
        THR(m_mesh->UnlockVertexBuffer());
    }
    operator Vertex *() const { return m_data; }

private:
    CComPtr<ID3DXMesh> m_mesh;
    Vertex *m_data;
};

//-----------------------------------------------------------------------------
// d3dframe_font
//
// CD3DFont doesn't provide accessors for information passed into the c'tor,
// so this little wrapper remembers them and provides accessors for them.
// This lets me use a stock d3dfont.cpp in the project.
//
class d3dframe_font : public CD3DFont
{
public:
    d3dframe_font(LPCTSTR name, DWORD height, DWORD flags)
        : m_name(name),
        m_height(height),
        m_flags(flags),
        CD3DFont(const_cast<LPTSTR>(name), height, flags)
    {
    }
    ~d3dframe_font()
    {
    }

    const tstring &name() const { return m_name; }
    DWORD height() const { return m_height; }
    DWORD flags() const { return m_flags; }

private:
    tstring m_name;
    DWORD m_height;
    DWORD m_flags;
};

//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the
//       generic functionality needed in all Direct3D samples. CMyD3DApplication
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
class CMyD3DApplication : public CD3DApplication
{
public:
    LRESULT MsgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
    CMyD3DApplication();
    virtual ~CMyD3DApplication() {}

protected:
    HRESULT OneTimeSceneInit();
    HRESULT InitDeviceObjects();
    HRESULT RestoreDeviceObjects();
    HRESULT InvalidateDeviceObjects();
    HRESULT DeleteDeviceObjects();
    HRESULT Render();
    HRESULT FrameMove();
    HRESULT FinalCleanup();

private:
    D3DCOLOR m_bg;

    bool m_show_stats;
    D3DCOLOR m_stats_fg;
    d3dframe_font *m_stats_font;

    bool m_show_d3dframe;
    D3DCOLOR m_d3dframe_fg;
    tstring m_d3dframe_display;
    d3dframe_font *m_d3dframe_font;
    float m_d3dframe_pos[2];

    bool m_show_mesh;
    D3DXCOLOR m_mesh_fg;
    tstring m_mesh_display;
    LOGFONT m_mesh_lf;
    CComPtr<ID3DXMesh> m_mesh;

    bool m_show_d3dx;
    D3DCOLOR m_d3dx_fg;
    tstring m_d3dx_display;
    LOGFONT m_d3dx_lf;
    CComPtr<ID3DXFont> m_d3dx_font;
    RECT m_d3dx_rect;

    //-----------------------------------------------------------------------
    // logical_y
    //
    // Return logical pixels in Y direction for font computations.
    //
    UINT logical_y() const
    {
        c_get_dc dc(m_hWnd);
        return ::GetDeviceCaps(dc, LOGPIXELSY);
    }

    LRESULT on_command(HWND window, WPARAM wp, LPARAM lp, bool &handled);

    void set_state();
    void set_view();
    void set_d3dx_rect();
    void set_d3dframe_pos();
    void change_text(tstring &display_text);
    void choose_font(d3dframe_font *&font);
    void choose_font(LOGFONT &lf, DWORD flags);
    void create_text();

    static INT_PTR CALLBACK
    change_text_proc(HWND dialog, UINT msg, WPARAM wp, LPARAM lp);
};

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI
WinMain(HINSTANCE hInst, HINSTANCE, LPSTR, INT)
{
    CMyD3DApplication d3dApp;

    if (FAILED(d3dApp.Create(hInst)))
        return 0;

    return d3dApp.Run();
}

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication()
    : m_bg(D3DCOLOR_XRGB(0, 0, 0)),

    m_show_stats(true),
    m_stats_fg(D3DCOLOR_XRGB(200, 200, 40)),
    m_stats_font(new d3dframe_font(_T("Arial"), 12, D3DFONT_BOLD)),

    // m_show_XXX initial values match menu check marks in resource file
    m_show_d3dframe(true),
    m_d3dframe_fg(D3DCOLOR_XRGB(200, 40, 100)),
    m_d3dframe_display(_T("Snappy Jackrabbit")),
    m_d3dframe_font(new d3dframe_font(_T("Arial"), 12, D3DFONT_BOLD)),

    m_show_mesh(true),
    m_mesh_fg(0.5f, 0.8f, 1.0f, 1.0f),
    m_mesh_display(_T("Quick Brown Fox")),
    m_mesh(),

    m_show_d3dx(true),
    m_d3dx_fg(D3DCOLOR_XRGB(40, 200, 40)),
    m_d3dx_display(_T("Lazy Yellow Dog")),
    m_d3dx_font()
{
    m_strWindowTitle    = _T("rt_Text");
    m_bUseDepthBuffer   = TRUE;

    c_get_dc dc(::GetDesktopWindow());
    const LOGFONT lf = {
        -MulDiv(INITIAL_POINT_SIZE, ::GetDeviceCaps(dc, LOGPIXELSY), 72),
        0, 0, 0, FW_NORMAL, false, false, false, DEFAULT_CHARSET,
        OUT_TT_ONLY_PRECIS, CLIP_DEFAULT_PRECIS, ANTIALIASED_QUALITY,
        DEFAULT_PITCH | FF_DONTCARE, _T("Arial")
    };
    m_d3dx_lf = lf;
    m_mesh_lf = lf;

    const RECT null = { 0 };
    m_d3dx_rect = null;
    m_d3dframe_pos[0] = m_d3dframe_pos[1] = 0;
}

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()
{
    return S_OK;
}

//-------------------------------------------------------------------------
// CMyD3DApplication::set_view
//
// Animate the eye point in a circle, but remain looking at the origin
//
void
CMyD3DApplication::set_view()
{
    D3DXVECTOR3 eye(cosf(m_fTime/D3DX_PI), sinf(m_fTime/D3DX_PI), -1);
    D3DXVECTOR3 at(0, 0, 0);
    D3DXVECTOR3 up(0, 1, 0);
    D3DXMATRIX view;
    ::D3DXMatrixLookAtLH(&view, &eye, &at, &up);
    THR(m_pd3dDevice->SetTransform(D3DTS_VIEW, &view));
}

//-------------------------------------------------------------------------
// CMyD3DApplication::set_d3dx_rect
//
// Move a box around a circle.  The box is the RECT used to confine the
// D3DX text.  The box is 1/4 the size of the screen in each dimension,
// making for 1/16th the area of the screen.
//
void
CMyD3DApplication::set_d3dx_rect()
{
    // dimensions of RECT containing text.
    const UINT box_w = m_d3dsdBackBuffer.Width/4;
    const UINT box_h = m_d3dsdBackBuffer.Height/4;

    // window geometry, allowing for twice margin on all sides
    const float aspect = box_w/float(box_h);
    const UINT margin = 20;
    const UINT height = m_d3dsdBackBuffer.Height - 2*margin - box_h;
    const UINT width = m_d3dsdBackBuffer.Width - 2*margin - box_w;

    // radius of circle centered at margin + {w,h}/2
    const UINT radius = ((aspect > 1) ? height : width)/2;

    // coordinates of RECT containing text.
    m_d3dx_rect.left = margin + width/2 +
        radius*cosf(m_fTime/D3DX_PI + 0.25f*D3DX_PI);
    m_d3dx_rect.top = margin + height/2 +
        radius*sinf(m_fTime/D3DX_PI + 0.25f*D3DX_PI);
    m_d3dx_rect.right = m_d3dx_rect.left + box_w;
    m_d3dx_rect.bottom = m_d3dx_rect.top + box_h;
}

//-----------------------------------------------------------------------------
// CMyD3DApplication::set_d3dframe_pos
//
// Move the D3DFrame text in a circle half a revolution out of phase with
// the D3DX text.
//
void
CMyD3DApplication::set_d3dframe_pos()
{
    const float aspect =
        m_d3dsdBackBuffer.Width/float(m_d3dsdBackBuffer.Height);
    const UINT margin = 2*m_d3dframe_font->height();
    // make a guess on the box width
    const UINT box_w = m_d3dframe_display.length()*m_d3dframe_font->height()/2;
    const UINT box_h = m_d3dframe_font->height();
    const UINT center_x = (m_d3dsdBackBuffer.Width - 3*margin - box_w)/2;
    const UINT center_y = (m_d3dsdBackBuffer.Height - 2*margin - box_h)/2;
    const UINT radius = (aspect > 1) ? center_y : center_x;

    m_d3dframe_pos[0] =
        margin + center_x + radius*cosf(D3DX_PI + 3*m_fTime/D3DX_PI);
    m_d3dframe_pos[1] =
        margin + center_y + radius*sinf(D3DX_PI + 3*m_fTime/D3DX_PI);
}

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove()
{
    if (m_show_mesh)
    {
        set_view();
    }

    if (m_show_d3dx)
    {
        set_d3dx_rect();
    }

    if (m_show_d3dframe)
    {
        set_d3dframe_pos();
    }

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
HRESULT
CMyD3DApplication::Render()
{
    THR(m_pd3dDevice->Clear(0L, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                            m_bg, 1.0f, 0L));
    THR(m_pd3dDevice->BeginScene());

    // display text in different forms
    if (m_show_mesh)
    {
        // We rely on the device having the proper state before drawing the
        // subset.  This is taken care of by set_state().
        THR(m_mesh->DrawSubset(0));
    }
    if (m_show_d3dx)
    {
        // Begin/End setup and restore device state for calls to
        // ID3DXFont::DrawText.  If we don't call these, DrawText will call
        // them for us.  We don't gain or lose anything by calling them
        // explicitly before DrawText since we only draw one string.  If
        // we were drawing many strings, we would gain by calling Begin/End
        // ourselves since the device state would be set once for all the
        // strings instead of being reset for each string.
        THR(m_d3dx_font->Begin());

        // The documentation incorrectly states that ID3DXFont::DrawText()
        // returns an HRESULT; it returns the count of characters drawn.
        // ID3DXFont::DrawText provides the behavior of the GDI function
        // DrawText() for a D3D device.  See DrawText in MSDN.
        //
        const UINT count =
            m_d3dx_font->DrawText(m_d3dx_display.c_str(), -1, &m_d3dx_rect,
                                  DT_CENTER | DT_VCENTER | DT_WORDBREAK,
                                  m_d3dx_fg);

        THR(m_d3dx_font->End());
    }
    if (m_show_d3dframe)
    {
        text(m_d3dframe_font, m_d3dframe_pos[0], m_d3dframe_pos[1],
             m_d3dframe_fg, m_d3dframe_display);
    }

    // Show frame rate and device statistics
    if (m_show_stats)
    {
        text(m_stats_font, 2,  0, m_stats_fg, m_strFrameStats);
        text(m_stats_font, 2, 20, m_stats_fg, m_strDeviceStats);
    }

    THR(m_pd3dDevice->EndScene());

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{
    // Restore the fonts
    m_d3dframe_font->InitDeviceObjects(m_pd3dDevice);
    m_stats_font->InitDeviceObjects(m_pd3dDevice);

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
    // Restore the fonts
    m_d3dframe_font->RestoreDeviceObjects();
    m_stats_font->RestoreDeviceObjects();

    create_text();

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{
    m_d3dframe_font->InvalidateDeviceObjects();
    m_stats_font->InvalidateDeviceObjects();
    m_d3dx_font = 0;
    m_mesh = 0;

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{
    m_d3dframe_font->DeleteDeviceObjects();
    m_stats_font->DeleteDeviceObjects();
    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
    destroy(m_d3dframe_font);
    destroy(m_stats_font);

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
LRESULT
CMyD3DApplication::MsgProc(HWND window, UINT msg, WPARAM wp, LPARAM lp)
{
    bool handled = false;
    LRESULT res = 0;

    // simple 'message map' style macro
#define RT_HANDLE_MSG(msg_, fn_) \
    case msg_: res = fn_(window, wp, lp, handled); break

    switch (msg)
    {
    RT_HANDLE_MSG(WM_COMMAND, on_command);
    }
#undef RT_HANDLE_MSG

    return handled ? res : CD3DApplication::MsgProc(window, msg, wp, lp);
}


//-----------------------------------------------------------------------------
// CMyD3DApplication::on_command
//
// WM_COMMAND handler: process toggles, font choosers and change text dialogs.
//
LRESULT
CMyD3DApplication::on_command(HWND window, WPARAM wp, LPARAM lp, bool &handled)
{
    const UINT control = LOWORD(wp);
    const HMENU menu = ::GetMenu(window);

    handled = true;
    switch (control)
    {
    // toggle show controls
#define RT_TOGGLE(id_, state_) \
    case id_: \
        toggle_menu(state_, menu, id_); \
        break
    RT_TOGGLE(IDM_SHOW_D3DFRAME, m_show_d3dframe);
    RT_TOGGLE(IDM_SHOW_MESH, m_show_mesh);
    RT_TOGGLE(IDM_SHOW_D3DX, m_show_d3dx);
    RT_TOGGLE(IDM_SHOW_STATISTICS, m_show_stats);
#undef RT_TOGGLE

    // choose fonts
    case IDM_MESH_FONT:
        choose_font(m_mesh_lf, CF_TTONLY);
        break;

    case IDM_D3DX_FONT:
        choose_font(m_d3dx_lf, CF_SCALABLEONLY);
        break;

    case IDM_D3DFRAME_FONT:
        choose_font(m_d3dframe_font);
        break;

    // change text
#define RT_CHANGE_TEXT(id_, state_) \
    case id_: \
        change_text(state_); \
        break
    RT_CHANGE_TEXT(IDM_MESH_TEXT, m_mesh_display);
    RT_CHANGE_TEXT(IDM_D3DX_TEXT, m_d3dx_display);
    RT_CHANGE_TEXT(IDM_D3DFRAME_TEXT, m_d3dframe_display);
#undef RT_CHANGE_TEXT

    // change colors: text and background
    case IDM_MESH_COLOR:
        m_mesh_fg = rt_choose_color(m_hWnd, m_mesh_fg);
        // need to reset material
        set_state();
        break;

#define RT_SET_OPAQUE(id_, state_) \
    case id_: \
        state_ = rt_choose_color(m_hWnd, state_) | D3DCOLOR_XRGB(0,0,0); \
        break
#define RT_SET_COLOR(id_, state_) \
    case id_: \
        state_ = rt_choose_color(m_hWnd, state_); \
        break
    RT_SET_COLOR(IDM_D3DX_COLOR, m_d3dx_fg);
    RT_SET_COLOR(IDM_D3DFRAME_COLOR, m_d3dframe_fg);
    RT_SET_OPAQUE(IDM_STATISTICS_COLOR, m_stats_fg);
    RT_SET_OPAQUE(IDM_BACKGROUND_COLOR, m_bg);
#undef RT_SET_COLOR
#undef RT_SET_OPAQUE

    default:
        // wasn't something we recognized, so we didn't handle it after all
        handled = false;
    }

    return 0;
}

//-----------------------------------------------------------------------------
// CMyD3DApplication::set_state
//
// We set the renderstates appropriate for the mesh.  The ID3DXFont class
// saves the existing state, sets what it needs and then restores the state
// before returning to the caller.
//
void
CMyD3DApplication::set_state()
{
    // setup device states to something reasonable for us
    const s_rs states[] =
    {
        D3DRS_CLIPPLANEENABLE, 0L,
        D3DRS_VERTEXBLEND, D3DVBF_DISABLE,
        D3DRS_FILLMODE, D3DFILL_SOLID,
        D3DRS_SHADEMODE, D3DSHADE_GOURAUD,
        D3DRS_CULLMODE, D3DCULL_CCW,
        D3DRS_FOGENABLE, FALSE,
        D3DRS_SPECULARENABLE, TRUE,
        D3DRS_LIGHTING, TRUE,
        D3DRS_AMBIENT, D3DCOLOR_XRGB(50, 50, 50),
        D3DRS_COLORVERTEX, FALSE,
        D3DRS_AMBIENTMATERIALSOURCE, D3DMCS_MATERIAL,
        D3DRS_DIFFUSEMATERIALSOURCE, D3DMCS_MATERIAL,
        D3DRS_SPECULARMATERIALSOURCE, D3DMCS_MATERIAL,
        D3DRS_EMISSIVEMATERIALSOURCE, D3DMCS_MATERIAL,
        D3DRS_ALPHATESTENABLE, FALSE,
        D3DRS_ZENABLE, D3DZB_TRUE,
        D3DRS_ZFUNC, D3DCMP_LESS,
        D3DRS_ZBIAS, 0,
        D3DRS_STENCILENABLE, FALSE,
        D3DRS_ALPHABLENDENABLE, FALSE,
        D3DRS_COLORWRITEENABLE, 0xF,
        D3DRS_DITHERENABLE, TRUE,
    };
    set_states(m_pd3dDevice, states, NUM_OF(states));

    // Set up a material
    D3DMATERIAL8 material;
    ::D3DUtil_InitMaterial(material, m_mesh_fg.r, m_mesh_fg.g, m_mesh_fg.b);
    THR(m_pd3dDevice->SetMaterial(&material));

    // Set up the light
    D3DLIGHT8 light;
    ::D3DUtil_InitLight(light, D3DLIGHT_DIRECTIONAL, 0.0f, -1.0f, 1.0f);
    THR(m_pd3dDevice->SetLight(0, &light));
    THR(m_pd3dDevice->LightEnable(0, TRUE));

    // figure out the bounding box of the mesh to adjust the projection
    // and mesh position.
    D3DXVECTOR3 min, max;
    {
        mesh_vb_lock<s_d3dx_vertex> vertices(m_mesh);
        THR(D3DXComputeBoundingBox(vertices, m_mesh->GetNumVertices(),
                                   m_mesh->GetFVF(), &min, &max));
    }

    // position the mesh centered around the origin.
    D3DXMATRIX world;
    ::D3DXMatrixTranslation(&world,
        -(min.x + max.x)/2, -(min.y + max.y)/2, -(min.z + max.z)/2);
    THR(m_pd3dDevice->SetTransform(D3DTS_WORLD, &world));

    // put the eye initially up in the corner at (x,y)=(1,0)=(cos(0),sin(0))
    set_view();

    // figure out a scaling factor that keeps the mesh in proportion,
    // but scales it to fit the aspect ratio of the window.
    const float back_aspect =
        m_d3dsdBackBuffer.Width / float(m_d3dsdBackBuffer.Height);
    const float width = max.x - min.x;
    const float height = max.y - min.y;
    const float mesh_aspect = width/height;

    // the 1.05 gives a 5% margin at all four edges
    float x_factor = 1.05f*width;
    float y_factor = 1.05f*height;

    // two cases defined by the relative aspect
    // ratios of the mesh and the back buffer.
    if (mesh_aspect > back_aspect)
    {
        y_factor = x_factor/back_aspect;
    }
    else
    {
        x_factor = y_factor/back_aspect;
    }

    // z_factor was heuristically determined, not mathematically
    const float z_factor = SQRT_2*max(x_factor, y_factor)/2;
    D3DXMATRIX proj;
    ::D3DXMatrixOrthoLH(&proj, x_factor, y_factor, -z_factor, z_factor);
    THR(m_pd3dDevice->SetTransform(D3DTS_PROJECTION, &proj));

    set_d3dx_rect();
    set_d3dframe_pos();
}

//-----------------------------------------------------------------------------
// CMyD3DApplication::create_text
//
// Create the mesh and the ID3DXFont based on current settings and reset
// device state.  (The matrices depend on the mesh bounding box.)
//
void
CMyD3DApplication::create_text()
{
    // release any existing.
    m_d3dx_font = 0;
    m_mesh = 0;

    // create ID3DXFont
    {
        HFONT font = ::CreateFontIndirect(&m_d3dx_lf);
        THR(::D3DXCreateFont(m_pd3dDevice, font, &m_d3dx_font));
        ::DeleteObject(font);
    }

    // create text as mesh
    {
        HFONT font = ::CreateFontIndirect(&m_mesh_lf);
        c_get_dc dc(m_hWnd);
        c_push_gdi<HFONT> pp(dc, font);
        THR(::D3DXCreateText(m_pd3dDevice, dc, m_mesh_display.c_str(),
                             0.001f, 0.25f, &m_mesh, NULL));
        ::DeleteObject(font);
    }

    set_state();
}

//-----------------------------------------------------------------------------
// CMyD3DApplication::choose_font
//
// Fill in boiler plate arguments for ::ChooseFont to select either the
// mesh LOGFONT or the ID3DXFont LOGFONT.  If the choice was confirmed,
// then copy the new LOGFONT to the result and recreate the text.
//
// CD3DFont has a different font selection mechanism, so I didn't bother
// to code a font selection for it.
//
void
CMyD3DApplication::choose_font(LOGFONT &result, DWORD flags)
{
    LOGFONT lf = result;
    CHOOSEFONT cf = { sizeof(cf) };
    cf.hwndOwner = m_hWnd;
    cf.lpLogFont = &lf;
    cf.iPointSize = INITIAL_POINT_SIZE;
    cf.Flags = CF_SCREENFONTS | CF_INITTOLOGFONTSTRUCT | flags;
    if (::ChooseFont(&cf))
    {
        result = lf;
        create_text();
    }
}

//-----------------------------------------------------------------------------
// CMyD3DApplication::change_text_proc
//
// Dialog procedure for the change text dialog.  Populate the edit control
// on WM_INITDIALOG and populate the text on WM_COMMAND IDOK.
//
INT_PTR CALLBACK
CMyD3DApplication::change_text_proc(HWND dialog, UINT msg,
                                    WPARAM wp, LPARAM lp)
{
    // HAZARD: this limits us to 1 dialog at a time.  ATL/WTL and MFC
    // frameworks provide better solutions for real applications.
    static tstring *text = NULL;

    HWND edit = ::GetDlgItem(dialog, IDC_EDIT_DISPLAY_TEXT);

    if (WM_INITDIALOG == msg)
    {
        text = reinterpret_cast<tstring *>(lp);
        ATLASSERT(edit);
        Edit_SetText(edit, text->c_str());
        return TRUE;
    }
    else if (WM_COMMAND == msg)
    {
        const UINT control = LOWORD(wp);
        if (IDOK == control)
        {
            // can't use tstring::c_str() here
            std::vector<TCHAR> value(Edit_GetTextLength(edit)+1);
            ATLASSERT(edit);
            Edit_GetText(edit, &value[0], value.size());
            *text = &value[0];

            ::EndDialog(dialog, IDOK);

            return TRUE;
        }
        else if (IDCANCEL == control)
        {
            ::EndDialog(dialog, IDCANCEL);
            return TRUE;
        }
    }

    return FALSE;
}

//-----------------------------------------------------------------------------
// CMyD3DApplication::change_text
//
// Launches the dialog to provide the user with a way to change the display
// text.  If the dialog was confirmed, then recreate the text.
//
void
CMyD3DApplication::change_text(tstring &display_text)
{
    if (IDOK ==
        DialogBoxParam(GetModuleHandle(NULL),
                       MAKEINTRESOURCE(IDD_CHANGE_TEXT), m_hWnd,
                       change_text_proc,
                       reinterpret_cast<LPARAM>(&display_text)))
    {
        create_text();
    }
}

//-----------------------------------------------------------------------------
// CMyD3DApplication::choose_font
//
// Choose a new CD3DFont.
//
void
CMyD3DApplication::choose_font(d3dframe_font *&font)
{
    // build a LOGFONT from the d3dframe_font info
    LOGFONT lf =
    {
        -MulDiv(font->height(), logical_y(), 72),
        0, 0, 0,
        (font->flags() & D3DFONT_BOLD) ? FW_BOLD : FW_NORMAL,
        (font->flags() & D3DFONT_ITALIC) ? TRUE : FALSE,
        FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS,
        CLIP_DEFAULT_PRECIS, ANTIALIASED_QUALITY, DEFAULT_PITCH
    };
    _tcsncpy(lf.lfFaceName, font->name().c_str(), min(80, LF_FACESIZE));

    // choose new font based on constructed LOGFONT
    CHOOSEFONT cf = { sizeof(cf) };
    cf.hwndOwner = m_hWnd;
    cf.lpLogFont = &lf;
    cf.Flags = CF_SCREENFONTS | CF_INITTOLOGFONTSTRUCT;
    if (::ChooseFont(&cf))
    {
        // nuke old font
        delete font;

        // figure out params for new font
        const DWORD height = MulDiv(abs(lf.lfHeight), 72, logical_y());
        const DWORD flags =
            (lf.lfWeight & FW_BOLD ? D3DFONT_BOLD : 0)
            | (lf.lfItalic ? D3DFONT_ITALIC : 0);

        // init new font
        font = new d3dframe_font(lf.lfFaceName, height, flags);
        THR(font->InitDeviceObjects(m_pd3dDevice));
        THR(font->RestoreDeviceObjects());
    }
}
